@@ -1,14 +1,15 @@
 /*
- * Copyright (c) 2005-2010 Douglas Gilbert.
+ * Copyright (c) 2005-2015 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
  */
 
-/* sg_pt_freebsd version 1.10 20100321 */
+/* sg_pt_freebsd version 1.13 20150511 */
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdarg.h>
 #include <string.h>
 #include <sys/types.h>
 #include <dirent.h>
@@ -71,6 +72,26 @@ struct sg_pt_base {
     struct sg_pt_freebsd_scsi impl;
 };
 
+#ifdef __GNUC__
+static int pr2ws(const char * fmt, ...)
+        __attribute__ ((format (printf, 1, 2)));
+#else
+static int pr2ws(const char * fmt, ...);
+#endif
+
+
+static int
+pr2ws(const char * fmt, ...)
+{
+    va_list args;
+    int n;
+
+    va_start(args, fmt);
+    n = vfprintf(sg_warnings_strm ? sg_warnings_strm : stderr, fmt, args);
+    va_end(args);
+    return n;
+}
+
 
 /* Returns >= 0 if successful. If error in Unix returns negated errno. */
 int
@@ -93,8 +114,6 @@ scsi_pt_open_flags(const char * device_name,
     struct cam_device* cam_dev;
     int k;
 
-    if (NULL == sg_warnings_strm)
-        sg_warnings_strm = stderr;
     // Search table for a free entry
     for (k = 0; k < FREEBSD_MAXDEV; k++)
         if (! devicetable[k])
@@ -104,8 +123,7 @@ scsi_pt_open_flags(const char * device_name,
     // of "file descriptors" already allocated.
     if (k == FREEBSD_MAXDEV) {
         if (verbose)
-            fprintf(sg_warnings_strm, "too many open file descriptors "
-                    "(%d)\n", FREEBSD_MAXDEV);
+            pr2ws("too many open file descriptors (%d)\n", FREEBSD_MAXDEV);
         errno = EMFILE;
         return -1;
     }
@@ -123,7 +141,7 @@ scsi_pt_open_flags(const char * device_name,
     if (cam_get_device(device_name, fdchan->devname, DEV_IDLEN,
                        &(fdchan->unitnum)) == -1) {
         if (verbose)
-            fprintf(sg_warnings_strm, "bad device name structure\n");
+            pr2ws("bad device name structure\n");
         errno = EINVAL;
         return -1;
     }
@@ -131,8 +149,7 @@ scsi_pt_open_flags(const char * device_name,
     if (! (cam_dev = cam_open_spec_device(fdchan->devname,
                                           fdchan->unitnum, O_RDWR, NULL))) {
         if (verbose)
-            fprintf(sg_warnings_strm, "cam_open_spec_device: %s\n",
-                    cam_errbuf);
+            pr2ws("cam_open_spec_device: %s\n", cam_errbuf);
         errno = EPERM;    /* permissions or no CAM */
         return -1;
     }
@@ -172,6 +189,12 @@ construct_scsi_pt_obj()
 {
     struct sg_pt_freebsd_scsi * ptp;
 
+    /* The following 2 lines are temporary. It is to avoid a NULL pointer
+     * crash when an old utility is used with a newer library built after
+     * the sg_warnings_strm cleanup */
+    if (NULL == sg_warnings_strm)
+        sg_warnings_strm = stderr;
+
     ptp = (struct sg_pt_freebsd_scsi *)
                 calloc(1, sizeof(struct sg_pt_freebsd_scsi));
     if (ptp) {
@@ -199,6 +222,8 @@ clear_scsi_pt_obj(struct sg_pt_base * vp)
     struct sg_pt_freebsd_scsi * ptp = &vp->impl;
 
     if (ptp) {
+        if (ptp->ccb)
+            cam_freeccb(ptp->ccb);
         memset(ptp, 0, sizeof(struct sg_pt_freebsd_scsi));
         ptp->dxfer_dir = CAM_DIR_NONE;
     }
@@ -296,9 +321,8 @@ set_scsi_pt_task_attr(struct sg_pt_base * vp,
 void
 set_scsi_pt_flags(struct sg_pt_base * objp, int flags)
 {
-    /* do nothing, suppress warnings */
-    objp = objp;
-    flags = flags;
+    if (objp) { ; }     /* unused, suppress warning */
+    if (flags) { ; }     /* unused, suppress warning */
 }
 
 /* Executes SCSI command (or at least forwards it to lower layers).
@@ -313,46 +337,47 @@ do_scsi_pt(struct sg_pt_base * vp, int device_fd, int time_secs, int verbose)
     union ccb *ccb;
     int len, timout_ms;
 
-    if (NULL == sg_warnings_strm)
-        sg_warnings_strm = stderr;
     ptp->os_err = 0;
     if (ptp->in_err) {
         if (verbose)
-            fprintf(sg_warnings_strm, "Replicated or unused set_scsi_pt...\n");
+            pr2ws("Replicated or unused set_scsi_pt...\n");
         return SCSI_PT_DO_BAD_PARAMS;
     }
     if (NULL == ptp->cdb) {
         if (verbose)
-            fprintf(sg_warnings_strm, "No command (cdb) given\n");
+            pr2ws("No command (cdb) given\n");
         return SCSI_PT_DO_BAD_PARAMS;
     }
 
     if ((fd < 0) || (fd >= FREEBSD_MAXDEV)) {
         if (verbose)
-            fprintf(sg_warnings_strm, "Bad file descriptor\n");
+            pr2ws("Bad file descriptor\n");
         ptp->os_err = ENODEV;
         return -ptp->os_err;
     }
     fdchan = devicetable[fd];
     if (NULL == fdchan) {
         if (verbose)
-            fprintf(sg_warnings_strm, "File descriptor closed??\n");
+            pr2ws("File descriptor closed??\n");
         ptp->os_err = ENODEV;
         return -ptp->os_err;
     }
     if (NULL == fdchan->cam_dev) {
         if (verbose)
-            fprintf(sg_warnings_strm, "No open CAM device\n");
+            pr2ws("No open CAM device\n");
         return SCSI_PT_DO_BAD_PARAMS;
     }
 
-    if (! (ccb = cam_getccb(fdchan->cam_dev))) {
-        if (verbose)
-            fprintf(sg_warnings_strm, "cam_getccb: failed\n");
-        ptp->os_err = ENOMEM;
-        return -ptp->os_err;
-    }
-    ptp->ccb = ccb;
+    if (NULL == ptp->ccb) {     /* re-use if we have one already */
+        if (! (ccb = cam_getccb(fdchan->cam_dev))) {
+            if (verbose)
+                pr2ws("cam_getccb: failed\n");
+            ptp->os_err = ENOMEM;
+            return -ptp->os_err;
+        }
+        ptp->ccb = ccb;
+    } else
+        ccb = ptp->ccb;
 
     // clear out structure, except for header that was filled in for us
     bzero(&(&ccb->ccb_h)[1],
@@ -393,8 +418,11 @@ do_scsi_pt(struct sg_pt_base * vp, int device_fd, int time_secs, int verbose)
 
         if ((SAM_STAT_CHECK_CONDITION == ptp->scsi_status) ||
             (SAM_STAT_COMMAND_TERMINATED == ptp->scsi_status)) {
-            len = ptp->sense_len - ptp->sense_resid;
-            if (len)
+            if (ptp->sense_resid > ptp->sense_len)
+                len = ptp->sense_len;   /* crazy; ignore sense_resid */
+            else
+                len = ptp->sense_len - ptp->sense_resid;
+            if (len > 0)
                 memcpy(ptp->sense, &(ccb->csio.sense_data), len);
         }
     } else
@@ -442,10 +470,11 @@ int
 get_scsi_pt_sense_len(const struct sg_pt_base * vp)
 {
     const struct sg_pt_freebsd_scsi * ptp = &vp->impl;
-    int len;
 
-    len = ptp->sense_len - ptp->sense_resid;
-    return (len > 0) ? len : 0;
+    if (ptp->sense_resid > ptp->sense_len)
+        return ptp->sense_len;  /* strange; ignore ptp->sense_resid */
+    else
+        return ptp->sense_len - ptp->sense_resid;
 }
 
 int
