@@ -1,18 +1,20 @@
 /*
- * Copyright (c) 2006-2012 Douglas Gilbert.
+ * Copyright (c) 2006-2015 Douglas Gilbert.
  * All rights reserved.
  * Use of this source code is governed by a BSD-style
  * license that can be found in the BSD_LICENSE file.
  */
 
-/* sg_pt_win32 version 1.14 20121026 */
+/* sg_pt_win32 version 1.16 20150511 */
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <stddef.h>
+#include <stdarg.h>
 #include <string.h>
 #include <errno.h>
 #include <ctype.h>
+#include <fcntl.h>
 
 #include "sg_pt.h"
 #include "sg_lib.h"
@@ -22,6 +24,13 @@
 #include "config.h"
 #endif
 
+#ifndef O_EXCL
+// #define O_EXCL 0x80  // cygwin ??
+// #define O_EXCL 0x80  // Linux
+#define O_EXCL 0x400    // mingw
+#warning "O_EXCL not defined"
+#endif
+
 /* Use the Microsoft SCSI Pass Through (SPT) interface. It has two
  * variants: "SPT" where data is double buffered; and "SPTD" where data
  * pointers to the user space are passed to the OS. Only Windows
@@ -56,9 +65,10 @@ struct sg_pt_handle {
     int bus;
     int target;
     int lun;
+    int verbose;        /* tunnel verbose through to scsi_pt_close_device */
 };
 
-struct sg_pt_handle handle_arr[MAX_OPEN_SIMULT];
+static struct sg_pt_handle handle_arr[MAX_OPEN_SIMULT];
 
 struct sg_pt_win32_scsi {
     unsigned char * dxferp;
@@ -90,6 +100,26 @@ static int spt_direct = 1;
 static int spt_direct = 0;
 #endif
 
+#ifdef __GNUC__
+static int pr2ws(const char * fmt, ...)
+        __attribute__ ((format (printf, 1, 2)));
+#else
+static int pr2ws(const char * fmt, ...);
+#endif
+
+
+static int
+pr2ws(const char * fmt, ...)
+{
+    va_list args;
+    int n;
+
+    va_start(args, fmt);
+    n = vfprintf(sg_warnings_strm ? sg_warnings_strm : stderr, fmt, args);
+    va_end(args);
+    return n;
+}
+
 
 /* Request SPT direct interface when state_direct is 1, state_direct set
  * to 0 for the SPT indirect interface. */
@@ -130,25 +160,21 @@ scsi_pt_open_device(const char * device_name, int read_only, int verbose)
  * form.
  */
 int
-scsi_pt_open_flags(const char * device_name,
-                   int flags __attribute__ ((unused)),
-                   int verbose)
+scsi_pt_open_flags(const char * device_name, int flags, int verbose)
 {
     int len, k, adapter_num, bus, target, lun, off, got_scsi_name;
-    int index, num, got_pd_name, pd_num;
+    int index, num, got_pd_name, pd_num, share_mode;
     struct sg_pt_handle * shp;
     char buff[8];
 
-    if (NULL == sg_warnings_strm)
-        sg_warnings_strm = stderr;
+    share_mode = (O_EXCL & flags) ? 0 : (FILE_SHARE_READ | FILE_SHARE_WRITE);
     /* lock */
     for (k = 0; k < MAX_OPEN_SIMULT; k++)
         if (0 == handle_arr[k].in_use)
             break;
     if (k == MAX_OPEN_SIMULT) {
         if (verbose)
-            fprintf(sg_warnings_strm, "too many open handles "
-                    "(%d)\n", MAX_OPEN_SIMULT);
+            pr2ws("too many open handles (%d)\n", MAX_OPEN_SIMULT);
         return -EMFILE;
     } else
         handle_arr[k].in_use = 1;
@@ -182,8 +208,8 @@ scsi_pt_open_flags(const char * device_name,
                              &adapter_num, &bus, &target, &lun);
                 if (num < 3) {
                     if (verbose)
-                        fprintf(sg_warnings_strm, "expected format like: "
-                                "'SCSI<port>:<bus>.<target>[.<lun>]'\n");
+                        pr2ws("expected format like: "
+                              "'SCSI<port>:<bus>.<target>[.<lun>]'\n");
                     shp->in_use = 0;
                     return -EINVAL;
                 }
@@ -194,6 +220,7 @@ scsi_pt_open_flags(const char * device_name,
     shp->bus = bus;
     shp->target = target;
     shp->lun = lun;
+    shp->verbose = verbose;
     memset(shp->adapter, 0, sizeof(shp->adapter));
     strncpy(shp->adapter, "\\\\.\\", 4);
     if (got_pd_name)
@@ -206,12 +233,11 @@ scsi_pt_open_flags(const char * device_name,
         snprintf(shp->adapter + 4, sizeof(shp->adapter) - 5, "%s",
                  device_name + off);
     shp->fh = CreateFile(shp->adapter, GENERIC_READ | GENERIC_WRITE,
-                         FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
-                         OPEN_EXISTING, 0, NULL);
+                         share_mode, NULL, OPEN_EXISTING, 0, NULL);
     if (shp->fh == INVALID_HANDLE_VALUE) {
         if (verbose)
-            fprintf(sg_warnings_strm, "Windows CreateFile error=%ld\n",
-                    GetLastError());
+            pr2ws("Windows CreateFile error=%u\n",
+                  (unsigned int)GetLastError());
         shp->in_use = 0;
         return -ENODEV;
     }
@@ -219,7 +245,11 @@ scsi_pt_open_flags(const char * device_name,
 }
 
 
-/* Returns 0 if successful. If error in Unix returns negated errno. */
+/* Returns 0 if successful. If device_id seems wild returns -ENODEV,
+ * other errors return 0. If CloseHandle() fails and verbose > 0 then
+ * outputs warning with value from GetLastError(). The verbose value
+ * defaults to zero and is potentially set from the most recent call
+ * to scsi_pt_open_device() or do_scsi_pt(). */
 int
 scsi_pt_close_device(int device_fd)
 {
@@ -231,12 +261,14 @@ scsi_pt_close_device(int device_fd)
     if ((index < 0) || (index >= WIN32_FDOFFSET))
         return -ENODEV;
     shp = handle_arr + index;
-    CloseHandle(shp->fh);
+    if ((! CloseHandle(shp->fh)) && shp->verbose)
+        pr2ws("Windows CloseHandle error=%u\n", (unsigned int)GetLastError());
     shp->bus = 0;
     shp->target = 0;
     shp->lun = 0;
     memset(shp->adapter, 0, sizeof(shp->adapter));
     shp->in_use = 0;
+    shp->verbose = 0;
     return 0;
 }
 
@@ -246,6 +278,12 @@ construct_scsi_pt_obj()
     struct sg_pt_win32_scsi * psp;
     struct sg_pt_base * vp = NULL;
 
+    /* The following 2 lines are temporary. It is to avoid a NULL pointer
+     * crash when an old utility is used with a newer library built after
+     * the sg_warnings_strm cleanup */
+    if (NULL == sg_warnings_strm)
+        sg_warnings_strm = stderr;
+
     psp = (struct sg_pt_win32_scsi *)calloc(sizeof(struct sg_pt_win32_scsi),
                                             1);
     if (psp) {
@@ -429,7 +467,7 @@ set_scsi_pt_flags(struct sg_pt_base * objp, int flags)
 /* Executes SCSI command (or at least forwards it to lower layers)
  * using direct interface. Clears os_err field prior to active call (whose
  * result may set it again). */
-int
+static int
 do_scsi_pt_direct(struct sg_pt_base * vp, int device_fd, int time_secs,
                   int verbose)
 {
@@ -437,36 +475,35 @@ do_scsi_pt_direct(struct sg_pt_base * vp, int device_fd, int time_secs,
     struct sg_pt_win32_scsi * psp = vp->implp;
     struct sg_pt_handle * shp;
     BOOL status;
-    ULONG returned;
+    DWORD returned;
 
-    if (NULL == sg_warnings_strm)
-        sg_warnings_strm = stderr;
     psp->os_err = 0;
     if (psp->in_err) {
         if (verbose)
-            fprintf(sg_warnings_strm, "Replicated or unused set_scsi_pt...\n");
+            pr2ws("Replicated or unused set_scsi_pt...\n");
         return SCSI_PT_DO_BAD_PARAMS;
     }
     if (0 == psp->swb_d.spt.CdbLength) {
         if (verbose)
-            fprintf(sg_warnings_strm, "No command (cdb) given\n");
+            pr2ws("No command (cdb) given\n");
         return SCSI_PT_DO_BAD_PARAMS;
     }
 
     index = device_fd - WIN32_FDOFFSET;
     if ((index < 0) || (index >= WIN32_FDOFFSET)) {
         if (verbose)
-            fprintf(sg_warnings_strm, "Bad file descriptor\n");
+            pr2ws("Bad file descriptor\n");
         psp->os_err = ENODEV;
         return -psp->os_err;
     }
     shp = handle_arr + index;
     if (0 == shp->in_use) {
         if (verbose)
-            fprintf(sg_warnings_strm, "File descriptor closed??\n");
+            pr2ws("File descriptor closed??\n");
         psp->os_err = ENODEV;
         return -psp->os_err;
     }
+    shp->verbose = verbose;
     psp->swb_d.spt.Length = sizeof (SCSI_PASS_THROUGH_DIRECT);
     psp->swb_d.spt.PathId = shp->bus;
     psp->swb_d.spt.TargetId = shp->target;
@@ -474,19 +511,20 @@ do_scsi_pt_direct(struct sg_pt_base * vp, int device_fd, int time_secs,
     psp->swb_d.spt.TimeOutValue = time_secs;
     psp->swb_d.spt.DataTransferLength = psp->dxfer_len;
     if (verbose > 4) {
-        fprintf(stderr, " spt_direct, adapter: %s  Length=%d ScsiStatus=%d "
-                "PathId=%d TargetId=%d Lun=%d\n", shp->adapter,
-                (int)psp->swb_d.spt.Length,
-                (int)psp->swb_d.spt.ScsiStatus, (int)psp->swb_d.spt.PathId,
-                (int)psp->swb_d.spt.TargetId, (int)psp->swb_d.spt.Lun);
-        fprintf(stderr, "    CdbLength=%d SenseInfoLength=%d DataIn=%d "
-                "DataTransferLength=%lu\n",
-                (int)psp->swb_d.spt.CdbLength,
-                (int)psp->swb_d.spt.SenseInfoLength,
-                (int)psp->swb_d.spt.DataIn,
-                psp->swb_d.spt.DataTransferLength);
-        fprintf(stderr, "    TimeOutValue=%lu SenseInfoOffset=%lu\n",
-                psp->swb_d.spt.TimeOutValue, psp->swb_d.spt.SenseInfoOffset);
+        pr2ws(" spt_direct, adapter: %s  Length=%d ScsiStatus=%d PathId=%d "
+              "TargetId=%d Lun=%d\n", shp->adapter,
+              (int)psp->swb_d.spt.Length, (int)psp->swb_d.spt.ScsiStatus,
+              (int)psp->swb_d.spt.PathId, (int)psp->swb_d.spt.TargetId,
+              (int)psp->swb_d.spt.Lun);
+        pr2ws("    CdbLength=%d SenseInfoLength=%d DataIn=%d "
+              "DataTransferLength=%u\n",
+              (int)psp->swb_d.spt.CdbLength,
+              (int)psp->swb_d.spt.SenseInfoLength,
+              (int)psp->swb_d.spt.DataIn,
+              (unsigned int)psp->swb_d.spt.DataTransferLength);
+        pr2ws("    TimeOutValue=%u SenseInfoOffset=%u\n",
+              (unsigned int)psp->swb_d.spt.TimeOutValue,
+              (unsigned int)psp->swb_d.spt.SenseInfoOffset);
     }
     psp->swb_d.spt.DataBuffer = psp->dxferp;
     status = DeviceIoControl(shp->fh, IOCTL_SCSI_PASS_THROUGH_DIRECT,
@@ -497,10 +535,12 @@ do_scsi_pt_direct(struct sg_pt_base * vp, int device_fd, int time_secs,
                             &returned,
                             NULL);
     if (! status) {
-        psp->transport_err = GetLastError();
+        unsigned int u;
+
+        u = (unsigned int)GetLastError();
         if (verbose)
-            fprintf(sg_warnings_strm, "Windows DeviceIoControl error=%d\n",
-                    psp->transport_err);
+            pr2ws("Windows DeviceIoControl error=%u\n", u);
+        psp->transport_err = (int)u;
         psp->os_err = EIO;
         return 0;       /* let app find transport error */
     }
@@ -531,51 +571,48 @@ do_scsi_pt_indirect(struct sg_pt_base * vp, int device_fd, int time_secs,
     struct sg_pt_win32_scsi * psp = vp->implp;
     struct sg_pt_handle * shp;
     BOOL status;
-    ULONG returned;
+    DWORD returned;
 
-    if (NULL == sg_warnings_strm)
-        sg_warnings_strm = stderr;
     psp->os_err = 0;
     if (psp->in_err) {
         if (verbose)
-            fprintf(sg_warnings_strm, "Replicated or unused "
-                    "set_scsi_pt...\n");
+            pr2ws("Replicated or unused set_scsi_pt...\n");
         return SCSI_PT_DO_BAD_PARAMS;
     }
     if (0 == psp->swb_i.spt.CdbLength) {
         if (verbose)
-            fprintf(sg_warnings_strm, "No command (cdb) given\n");
+            pr2ws("No command (cdb) given\n");
         return SCSI_PT_DO_BAD_PARAMS;
     }
 
     index = device_fd - WIN32_FDOFFSET;
     if ((index < 0) || (index >= WIN32_FDOFFSET)) {
         if (verbose)
-            fprintf(sg_warnings_strm, "Bad file descriptor\n");
+            pr2ws("Bad file descriptor\n");
         psp->os_err = ENODEV;
         return -psp->os_err;
     }
     shp = handle_arr + index;
     if (0 == shp->in_use) {
         if (verbose)
-            fprintf(sg_warnings_strm, "File descriptor closed??\n");
+            pr2ws("File descriptor closed??\n");
         psp->os_err = ENODEV;
         return -psp->os_err;
     }
+    shp->verbose = verbose;
     if (psp->dxfer_len > (int)sizeof(psp->swb_i.ucDataBuf)) {
         int extra = psp->dxfer_len - (int)sizeof(psp->swb_i.ucDataBuf);
         struct sg_pt_win32_scsi * epsp;
 
         if (verbose > 4)
-            fprintf(sg_warnings_strm, "spt_indirect: dxfer_len (%d) too "
-                    "large for initial data\n  buffer (%d bytes), try "
-                    "enlarging\n", psp->dxfer_len,
-                    sizeof(psp->swb_i.ucDataBuf));
+            pr2ws("spt_indirect: dxfer_len (%d) too large for initial data\n"
+                  "  buffer (%d bytes), try enlarging\n", psp->dxfer_len,
+                  (int)sizeof(psp->swb_i.ucDataBuf));
         epsp = (struct sg_pt_win32_scsi *)
                calloc(sizeof(struct sg_pt_win32_scsi) + extra, 1);
         if (NULL == epsp) {
-            fprintf(sg_warnings_strm, "do_scsi_pt: failed to enlarge data "
-                    "buffer to %d bytes\n", psp->dxfer_len);
+            pr2ws("do_scsi_pt: failed to enlarge data buffer to %d bytes\n",
+                  psp->dxfer_len);
             psp->os_err = ENOMEM;
             return -psp->os_err;
         }
@@ -593,21 +630,22 @@ do_scsi_pt_indirect(struct sg_pt_base * vp, int device_fd, int time_secs,
     psp->swb_i.spt.TimeOutValue = time_secs;
     psp->swb_i.spt.DataTransferLength = psp->dxfer_len;
     if (verbose > 4) {
-        fprintf(stderr, " spt_indirect, adapter: %s  Length=%d ScsiStatus=%d "
-                "PathId=%d TargetId=%d Lun=%d\n", shp->adapter,
-                (int)psp->swb_i.spt.Length,
-                (int)psp->swb_i.spt.ScsiStatus, (int)psp->swb_i.spt.PathId,
-                (int)psp->swb_i.spt.TargetId, (int)psp->swb_i.spt.Lun);
-        fprintf(stderr, "    CdbLength=%d SenseInfoLength=%d DataIn=%d "
-                "DataTransferLength=%lu\n",
-                (int)psp->swb_i.spt.CdbLength,
-                (int)psp->swb_i.spt.SenseInfoLength,
-                (int)psp->swb_i.spt.DataIn,
-                psp->swb_i.spt.DataTransferLength);
-        fprintf(stderr, "    TimeOutValue=%lu DataBufferOffset=%lu "
-                "SenseInfoOffset=%lu\n", psp->swb_i.spt.TimeOutValue,
-                psp->swb_i.spt.DataBufferOffset,
-                psp->swb_i.spt.SenseInfoOffset);
+        pr2ws(" spt_indirect, adapter: %s  Length=%d ScsiStatus=%d PathId=%d "
+              "TargetId=%d Lun=%d\n", shp->adapter,
+              (int)psp->swb_i.spt.Length, (int)psp->swb_i.spt.ScsiStatus,
+              (int)psp->swb_i.spt.PathId, (int)psp->swb_i.spt.TargetId,
+              (int)psp->swb_i.spt.Lun);
+        pr2ws("    CdbLength=%d SenseInfoLength=%d DataIn=%d "
+              "DataTransferLength=%u\n",
+              (int)psp->swb_i.spt.CdbLength,
+              (int)psp->swb_i.spt.SenseInfoLength,
+              (int)psp->swb_i.spt.DataIn,
+              (unsigned int)psp->swb_i.spt.DataTransferLength);
+        pr2ws("    TimeOutValue=%u DataBufferOffset=%u "
+              "SenseInfoOffset=%u\n",
+              (unsigned int)psp->swb_i.spt.TimeOutValue,
+              (unsigned int)psp->swb_i.spt.DataBufferOffset,
+              (unsigned int)psp->swb_i.spt.SenseInfoOffset);
     }
     if ((psp->dxfer_len > 0) &&
         (SCSI_IOCTL_DATA_OUT == psp->swb_i.spt.DataIn))
@@ -620,10 +658,12 @@ do_scsi_pt_indirect(struct sg_pt_base * vp, int device_fd, int time_secs,
                             &returned,
                             NULL);
     if (! status) {
-        psp->transport_err = GetLastError();
+        unsigned int u;
+
+        u = (unsigned int)GetLastError();
         if (verbose)
-            fprintf(sg_warnings_strm, "Windows DeviceIoControl error=%d\n",
-                    psp->transport_err);
+            pr2ws("Windows DeviceIoControl error=%u\n", u);
+        psp->transport_err = (int)u;
         psp->os_err = EIO;
         return 0;       /* let app find transport error */
     }
