@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2006-2012 Douglas Gilbert.
+ * Copyright (c) 2006-2014 Douglas Gilbert.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -41,7 +41,16 @@
 
 #include "sg_lib.h"
 
+#ifdef _WIN32_WINNT
+ #if _WIN32_WINNT < 0x0602
+ #undef _WIN32_WINNT
+ #define _WIN32_WINNT 0x0602
+ #endif
+#else
 #define _WIN32_WINNT 0x0602
+/* claim its W8 */
+#endif
+
 #include "sg_pt_win32.h"
 
 /*
@@ -155,7 +164,7 @@ union STORAGE_DEVICE_DESCRIPTOR_DATA {
 
 union STORAGE_DEVICE_UID_DATA {
     STORAGE_DEVICE_UNIQUE_IDENTIFIER desc;
-    char raw[512];
+    char raw[1060];
 };
 
 struct storage_elem {
@@ -265,20 +274,20 @@ query_dev_property(HANDLE hdevice,
         if (verbose > 2) {
             err = GetLastError();
             fprintf(stderr, "  IOCTL_STORAGE_QUERY_PROPERTY(Devprop) failed, "
-                    "Error=%ld %s\n", err, get_err_str(err, sizeof(b), b));
+                    "Error=%u %s\n", (unsigned int)err,
+                    get_err_str(err, sizeof(b), b));
         }
         return -ENOSYS;
     }
 
     if (verbose > 3)
         fprintf(stderr, "  IOCTL_STORAGE_QUERY_PROPERTY(DevProp) "
-                "num_out=%ld\n", num_out);
+                "num_out=%u\n", (unsigned int)num_out);
     return 0;
 }
 
 static int
-query_dev_uid(HANDLE hdevice,
-                    union STORAGE_DEVICE_UID_DATA * data)
+query_dev_uid(HANDLE hdevice, union STORAGE_DEVICE_UID_DATA * data)
 {
     DWORD num_out, err;
     char b[256];
@@ -286,22 +295,42 @@ query_dev_uid(HANDLE hdevice,
                                     PropertyStandardQuery, {0} };
 
     memset(data, 0, sizeof(*data));
+    num_out = 0;
+    query.QueryType = PropertyExistsQuery;
+    if (! DeviceIoControl(hdevice, IOCTL_STORAGE_QUERY_PROPERTY,
+                          &query, sizeof(query), NULL, 0, &num_out, NULL)) {
+        if (verbose > 2) {
+            err = GetLastError();
+            fprintf(stderr, "  IOCTL_STORAGE_QUERY_PROPERTY(DevUid(exists)) "
+                    "failed, Error=%u %s\n", (unsigned int)err,
+                    get_err_str(err, sizeof(b), b));
+        }
+        if (verbose > 3)
+            fprintf(stderr, "      num_out=%u\n", (unsigned int)num_out);
+        /* interpret any error to mean this property doesn't exist */
+        return 0;
+    }
+
+    query.QueryType = PropertyStandardQuery;
     if (! DeviceIoControl(hdevice, IOCTL_STORAGE_QUERY_PROPERTY,
                           &query, sizeof(query), data, sizeof(*data),
                           &num_out, NULL)) {
         if (verbose > 2) {
             err = GetLastError();
             fprintf(stderr, "  IOCTL_STORAGE_QUERY_PROPERTY(DevUid) failed, "
-                    "Error=%ld %s\n", err, get_err_str(err, sizeof(b), b));
+                    "Error=%u %s\n", (unsigned int)err,
+                    get_err_str(err, sizeof(b), b));
         }
         return -ENOSYS;
     }
     if (verbose > 3)
-        fprintf(stderr, "  IOCTL_STORAGE_QUERY_PROPERTY(DevUid) num_out=%ld\n",
-                num_out);
+        fprintf(stderr, "  IOCTL_STORAGE_QUERY_PROPERTY(DevUid) num_out=%u\n",
+                (unsigned int)num_out);
     return 0;
 }
 
+/* Updates storage_arr based on sep. Returns 1 if update occurred, 0 if
+ * no update occured. */
 static int
 check_devices(const struct storage_elem * sep)
 {
@@ -311,7 +340,18 @@ check_devices(const struct storage_elem * sep)
     for (k = 0; k < next_unused_elem; ++k, ++sarr) {
         if ('\0' == sarr->name[0])
             continue;
-        if (sep->qp_descriptor_valid && sarr->qp_descriptor_valid) {
+        if (sep->qp_uid_valid && sarr->qp_uid_valid) {
+            if (0 == memcmp(&sep->qp_uid, &sarr->qp_uid,
+                            sizeof(sep->qp_uid))) {
+                for (j = 0; j < (int)sizeof(sep->volume_letters); ++j) {
+                    if ('\0' == sarr->volume_letters[j]) {
+                        sarr->volume_letters[j] = sep->name[0];
+                        break;
+                    }
+                }
+                return 1;
+            }
+        } else if (sep->qp_descriptor_valid && sarr->qp_descriptor_valid) {
             if (0 == memcmp(&sep->qp_descriptor, &sarr->qp_descriptor,
                             sizeof(sep->qp_descriptor))) {
                 for (j = 0; j < (int)sizeof(sep->volume_letters); ++j) {
@@ -323,7 +363,6 @@ check_devices(const struct storage_elem * sep)
                 return 1;
             }
         }
-        // should do uid check here (probably before descriptor compare)
     }
     return 0;
 }
@@ -339,7 +378,7 @@ enum_scsi_adapters(void)
     BYTE bus;
     BOOL success;
     char adapter_name[64];
-    char inqDataBuff[2048];
+    char inqDataBuff[8192];
     PSCSI_ADAPTER_BUS_INFO  ai;
     char b[256];
 
@@ -352,7 +391,7 @@ enum_scsi_adapters(void)
             hole_count = 0;
             success = DeviceIoControl(fh, IOCTL_SCSI_GET_INQUIRY_DATA,
                                       NULL, 0, inqDataBuff,
-                                      sizeof(inqDataBuff), &dummy, FALSE);
+                                      sizeof(inqDataBuff), &dummy, NULL);
             if (success) {
                 PSCSI_BUS_DATA pbd;
                 PSCSI_INQUIRY_DATA pid;
@@ -386,16 +425,20 @@ enum_scsi_adapters(void)
             } else {
                 err = GetLastError();
                 fprintf(stderr, "%s: IOCTL_SCSI_GET_INQUIRY_DATA failed "
-                        "err=%lu\n\t%s",
-                        adapter_name, err, get_err_str(err, sizeof(b), b));
+                        "err=%u\n\t%s",
+                        adapter_name, (unsigned int)err,
+                        get_err_str(err, sizeof(b), b));
             }
             CloseHandle(fh);
         } else {
-            if (verbose > 3) {
-                err = GetLastError();
-                fprintf(stderr, "%s: CreateFile failed err=%lu\n\t%s",
-                        adapter_name, err, get_err_str(err, sizeof(b), b));
-            }
+            err = GetLastError();
+            if (ERROR_SHARING_VIOLATION == err)
+                fprintf(stderr, "%s: in use by other process (sharing "
+                        "violation [34])\n", adapter_name);
+            else if (verbose > 3)
+                fprintf(stderr, "%s: CreateFile failed err=%u\n\t%s",
+                        adapter_name, (unsigned int)err,
+                        get_err_str(err, sizeof(b), b));
             if (++hole_count >= MAX_HOLE_COUNT)
                 break;
         }
@@ -412,7 +455,7 @@ enum_volumes(char letter)
     struct storage_elem tmp_se;
 
     if (verbose > 2)
-        fprintf(stderr, "%s: enter\n", __FUNCTION__ );
+        fprintf(stderr, "%s: enter\n", __func__);
     for (k = 0; k < 24; ++k) {
         memset(&tmp_se, 0, sizeof(tmp_se));
         snprintf(adapter_name, sizeof (adapter_name), "\\\\.\\%c:", 'C' + k);
@@ -422,14 +465,12 @@ enum_volumes(char letter)
                         OPEN_EXISTING, 0, NULL);
         if (fh != INVALID_HANDLE_VALUE) {
             if (query_dev_property(fh, &tmp_se.qp_descriptor) < 0)
-                fprintf(stderr, "%s: query_dev_property failed\n",
-                        __FUNCTION__ );
+                fprintf(stderr, "%s: query_dev_property failed\n", __func__);
             else
                 tmp_se.qp_descriptor_valid = 1;
             if (query_dev_uid(fh, &tmp_se.qp_uid) < 0) {
                 if (verbose > 2)
-                    fprintf(stderr, "%s: query_dev_uid failed\n",
-                            __FUNCTION__ );
+                    fprintf(stderr, "%s: query_dev_uid failed\n", __func__);
             } else
                 tmp_se.qp_uid_valid = 1;
             if (('\0' == letter) || (letter == tmp_se.name[0]))
@@ -452,7 +493,7 @@ enum_pds(void)
     struct storage_elem tmp_se;
 
     if (verbose > 2)
-        fprintf(stderr, "%s: enter\n", __FUNCTION__ );
+        fprintf(stderr, "%s: enter\n", __func__);
     for (k = 0; k < MAX_PHYSICALDRIVE_NUM; ++k) {
         memset(&tmp_se, 0, sizeof(tmp_se));
         snprintf(adapter_name, sizeof (adapter_name),
@@ -463,25 +504,26 @@ enum_pds(void)
                         OPEN_EXISTING, 0, NULL);
         if (fh != INVALID_HANDLE_VALUE) {
             if (query_dev_property(fh, &tmp_se.qp_descriptor) < 0)
-                fprintf(stderr, "%s: query_dev_property failed\n",
-                        __FUNCTION__ );
+                fprintf(stderr, "%s: query_dev_property failed\n", __func__);
             else
                 tmp_se.qp_descriptor_valid = 1;
             if (query_dev_uid(fh, &tmp_se.qp_uid) < 0) {
                 if (verbose > 2)
-                    fprintf(stderr, "%s: query_dev_uid failed\n",
-                            __FUNCTION__ );
+                    fprintf(stderr, "%s: query_dev_uid failed\n", __func__);
             } else
                 tmp_se.qp_uid_valid = 1;
             hole_count = 0;
             memcpy(&storage_arr[next_unused_elem++], &tmp_se, sizeof(tmp_se));
             CloseHandle(fh);
         } else {
-            if (verbose > 3) {
-                err = GetLastError();
-                fprintf(stderr, "%s: CreateFile failed err=%lu\n\t%s",
-                        adapter_name, err, get_err_str(err, sizeof(b), b));
-            }
+            err = GetLastError();
+            if (ERROR_SHARING_VIOLATION == err)
+                fprintf(stderr, "%s: in use by other process (sharing "
+                        "violation [34])\n", adapter_name);
+            else if (verbose > 3)
+                fprintf(stderr, "%s: CreateFile failed err=%u\n\t%s",
+                        adapter_name, (unsigned int)err,
+                        get_err_str(err, sizeof(b), b));
             if (++hole_count >= MAX_HOLE_COUNT)
                 break;
         }
@@ -501,7 +543,7 @@ enum_cdroms(void)
     struct storage_elem tmp_se;
 
     if (verbose > 2)
-        fprintf(stderr, "%s: enter\n", __FUNCTION__ );
+        fprintf(stderr, "%s: enter\n", __func__);
     for (k = 0; k < MAX_CDROM_NUM; ++k) {
         memset(&tmp_se, 0, sizeof(tmp_se));
         snprintf(adapter_name, sizeof (adapter_name), "\\\\.\\CDROM%d", k);
@@ -511,25 +553,26 @@ enum_cdroms(void)
                         OPEN_EXISTING, 0, NULL);
         if (fh != INVALID_HANDLE_VALUE) {
             if (query_dev_property(fh, &tmp_se.qp_descriptor) < 0)
-                fprintf(stderr, "%s: query_dev_property failed\n",
-                        __FUNCTION__ );
+                fprintf(stderr, "%s: query_dev_property failed\n", __func__);
             else
                 tmp_se.qp_descriptor_valid = 1;
             if (query_dev_uid(fh, &tmp_se.qp_uid) < 0) {
                 if (verbose > 2)
-                    fprintf(stderr, "%s: query_dev_uid failed\n",
-                            __FUNCTION__ );
+                    fprintf(stderr, "%s: query_dev_uid failed\n", __func__);
             } else
                 tmp_se.qp_uid_valid = 1;
             hole_count = 0;
             memcpy(&storage_arr[next_unused_elem++], &tmp_se, sizeof(tmp_se));
             CloseHandle(fh);
         } else {
-            if (verbose > 3) {
-                err = GetLastError();
-                fprintf(stderr, "%s: CreateFile failed err=%lu\n\t%s",
-                        adapter_name, err, get_err_str(err, sizeof(b), b));
-            }
+            err = GetLastError();
+            if (ERROR_SHARING_VIOLATION == err)
+                fprintf(stderr, "%s: in use by other process (sharing "
+                        "violation [34])\n", adapter_name);
+            else if (verbose > 3)
+                fprintf(stderr, "%s: CreateFile failed err=%u\n\t%s",
+                        adapter_name, (unsigned int)err,
+                        get_err_str(err, sizeof(b), b));
             if (++hole_count >= MAX_HOLE_COUNT)
                 break;
         }
@@ -549,7 +592,7 @@ enum_tapes(void)
     struct storage_elem tmp_se;
 
     if (verbose > 2)
-        fprintf(stderr, "%s: enter\n", __FUNCTION__ );
+        fprintf(stderr, "%s: enter\n", __func__);
     for (k = 0; k < MAX_TAPE_NUM; ++k) {
         memset(&tmp_se, 0, sizeof(tmp_se));
         snprintf(adapter_name, sizeof (adapter_name), "\\\\.\\TAPE%d", k);
@@ -559,25 +602,26 @@ enum_tapes(void)
                         OPEN_EXISTING, 0, NULL);
         if (fh != INVALID_HANDLE_VALUE) {
             if (query_dev_property(fh, &tmp_se.qp_descriptor) < 0)
-                fprintf(stderr, "%s: query_dev_property failed\n",
-                        __FUNCTION__ );
+                fprintf(stderr, "%s: query_dev_property failed\n", __func__);
             else
                 tmp_se.qp_descriptor_valid = 1;
             if (query_dev_uid(fh, &tmp_se.qp_uid) < 0) {
                 if (verbose > 2)
-                    fprintf(stderr, "%s: query_dev_uid failed\n",
-                            __FUNCTION__ );
+                    fprintf(stderr, "%s: query_dev_uid failed\n", __func__);
             } else
                 tmp_se.qp_uid_valid = 1;
             hole_count = 0;
             memcpy(&storage_arr[next_unused_elem++], &tmp_se, sizeof(tmp_se));
             CloseHandle(fh);
         } else {
-            if (verbose > 3) {
-                err = GetLastError();
-                fprintf(stderr, "%s: CreateFile failed err=%lu\n\t%s",
-                        adapter_name, err, get_err_str(err, sizeof(b), b));
-            }
+            err = GetLastError();
+            if (ERROR_SHARING_VIOLATION == err)
+                fprintf(stderr, "%s: in use by other process (sharing "
+                        "violation [34])\n", adapter_name);
+            else if (verbose > 3)
+                fprintf(stderr, "%s: CreateFile failed err=%u\n\t%s",
+                        adapter_name, (unsigned int)err,
+                        get_err_str(err, sizeof(b), b));
             if (++hole_count >= MAX_HOLE_COUNT)
                 break;
         }
@@ -641,9 +685,13 @@ do_wscan(char letter, int show_bt, int scsi_scan)
                     printf("%s", sp->qp_descriptor.raw + j);
                 printf("\n");
                 if (verbose > 2)
-                    dStrHex(sp->qp_descriptor.raw, 144, 0);
+                    dStrHexErr(sp->qp_descriptor.raw, 144, 0);
             } else
                 printf("\n");
+            if ((verbose > 3) && sp->qp_uid_valid) {
+                printf("  UID valid, in hex:\n");
+                dStrHexErr(sp->qp_uid.raw, sizeof(sp->qp_uid.raw), 1);
+            }
         }
     }
 
